@@ -45,6 +45,8 @@ static void gs_window_init       (GSWindow      *window);
 static void gs_window_finalize   (GObject       *object);
 
 static gboolean popup_dialog_idle (GSWindow *window);
+static void gs_window_dialog_finish (GSWindow *window);
+static void remove_command_watches (GSWindow *window);
 
 enum {
         DIALOG_RESPONSE_CANCEL,
@@ -91,6 +93,8 @@ struct GSWindowPrivate
         gint       lock_pid;
         gint       lock_watch_id;
         gint       dialog_response;
+        gboolean   dialog_quit_requested;
+        gboolean   dialog_shake_in_progress;
 
         gint       keyboard_pid;
         gint       keyboard_watch_id;
@@ -925,6 +929,21 @@ gs_window_destroy (GSWindow *window)
 {
         g_return_if_fail (GS_IS_WINDOW (window));
 
+        if (window->priv->lock_pid > 0) {
+                gs_window_dialog_finish (window);
+        }
+
+        remove_popup_dialog_idle (window);
+        remove_command_watches (window);
+        remove_watchdog_timer (window);
+
+        if (window->priv->lock_box != NULL) {
+                gtk_container_remove (GTK_CONTAINER (window->priv->vbox), GTK_WIDGET (window->priv->lock_box));
+                window->priv->lock_box = NULL;
+
+                g_signal_emit (window, signals [DIALOG_DOWN], 0);
+        }
+
         gtk_widget_destroy (GTK_WIDGET (window));
 }
 
@@ -970,6 +989,7 @@ error_watch (GIOChannel   *source,
                 case G_IO_STATUS_ERROR:
                         finished = TRUE;
                         gs_debug ("Error reading from child: %s\n", error->message);
+                        g_error_free (error);
                         return FALSE;
                 case G_IO_STATUS_AGAIN:
                 default:
@@ -1282,6 +1302,7 @@ keyboard_command_watch (GIOChannel   *source,
                 case G_IO_STATUS_ERROR:
                         finished = TRUE;
                         gs_debug ("Error reading from child: %s\n", error->message);
+                        g_error_free (error);
                         return FALSE;
                 case G_IO_STATUS_AGAIN:
                 default:
@@ -1381,6 +1402,16 @@ gs_window_dialog_finish (GSWindow *window)
         remove_key_events (window);
 }
 
+static void
+maybe_kill_dialog (GSWindow *window)
+{
+        if (!window->priv->dialog_shake_in_progress
+            && window->priv->dialog_quit_requested
+            && window->priv->lock_pid > 0) {
+                kill (window->priv->lock_pid, SIGTERM);
+        }
+}
+
 /* very rudimentary animation for indicating an auth failure */
 static void
 shake_dialog (GSWindow *window)
@@ -1389,6 +1420,8 @@ shake_dialog (GSWindow *window)
         guint left;
         guint right;
 
+        window->priv->dialog_shake_in_progress = TRUE;
+
         for (i = 0; i < 9; i++) {
                 if (i % 2 == 0) {
                         left = 30;
@@ -1413,6 +1446,9 @@ shake_dialog (GSWindow *window)
 
                 g_usleep (10000);
         }
+
+        window->priv->dialog_shake_in_progress = FALSE;
+        maybe_kill_dialog (window);
 }
 
 static gboolean
@@ -1455,6 +1491,10 @@ lock_command_watch (GIOChannel   *source,
                                         window->priv->dialog_response = DIALOG_RESPONSE_CANCEL;
                                 }
                                 finished = TRUE;
+                        } else if (strstr (line, "REQUEST QUIT") != NULL) {
+                                gs_debug ("Got request for quit");
+                                window->priv->dialog_quit_requested = TRUE;
+                                maybe_kill_dialog (window);
                         }
                         break;
                 case G_IO_STATUS_EOF:
@@ -1463,6 +1503,7 @@ lock_command_watch (GIOChannel   *source,
                 case G_IO_STATUS_ERROR:
                         finished = TRUE;
                         gs_debug ("Error reading from child: %s\n", error->message);
+                        g_error_free (error);
                         return FALSE;
                 case G_IO_STATUS_AGAIN:
                 default:
@@ -1567,6 +1608,9 @@ popup_dialog_idle (GSWindow *window)
 
         set_invisible_cursor (GTK_WIDGET (window)->window, FALSE);
 
+        window->priv->dialog_quit_requested = FALSE;
+        window->priv->dialog_shake_in_progress = FALSE;
+
         result = spawn_on_window (window,
                                   command->str,
                                   &window->priv->lock_pid,
